/******************************************************************************
                       C++ Implementation: TriangleReorderImpl
                            -------------------
    description       :
    begin             : Mon Mar 28 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#include "TriangleReorder.h"

#include <stdio.h>
#include <assert.h>
#include <set>
#include <numeric>
#include <OpenMesh/Core/Mesh/PolyConnectivity.hh>
#include <OpenMesh/Core/Utils/AutoPropertyHandleT.hh>

#include "VertexCache.h"

namespace
{
  typedef OpenMesh::VertexHandle            VertexHandle;
  typedef OpenMesh::FaceHandle              FaceHandle;
  typedef OpenMesh::EdgeHandle              EdgeHandle;
  typedef OpenMesh::HalfedgeHandle          HalfedgeHandle;
}

namespace tr
{

class TriangleReorderImpl
{
private:
  typedef OpenMesh::PolyConnectivity        Connectivity;
  typedef OpenMesh::AutoPropertyHandleT< Connectivity, OpenMesh::FPropHandleT<int> > 
                                            OrderFPropHandle;
  typedef std::vector<OpenMesh::VertexHandle> VertexHandleArray;                        
  typedef std::set<FaceHandle>              FaceHandleSet;
  
private:
  Triangle*                                 tris_;
  uint                                      n_tris_;
  uint                                      n_vertices_;
  
  Connectivity*                             conn_;
  OrderFPropHandle                          order_fpph_;
  uint                                      next_id_;
  VertexHandleArray                         avail_vhs_;
  VertexCache*                              vertex_cache_;
  std::vector<Triangle>                     reord_tris_;
  std::vector<uint>                         n_visits_;
  //holds unordered faces with at least one vertex in the vertex_cache_
  FaceHandleSet                             cand_fhs_;
  
private:
  void                                      visit_vertex(int _vertex_index);
  void                                      visit_face(int _face_index);
  /// second version of the above functions which updates the cand_fhs_ set  
  void                                      visit_vertex_ext(int _vertex_index);
  void                                      visit_face_ext(int _face_index);
  void                                      order_face(FaceHandle _fh);
  
  void                                      build_connectivity();
  void                                      build_avail_vhs();
  FaceHandle                                find_next_face();
  
  uint                                      calc_face_score(FaceHandle _fh) const;
  double                                    calc_face_score_ext(FaceHandle _fh) const;
  FaceHandle                                find_best_adjacent_face(VertexHandle _vh) const;
  

public:
  TriangleReorderImpl(Triangle* _tris, uint _n_tris, uint _n_vertices, uint _cache_size);
  ~TriangleReorderImpl();
  
  uint                                      calc_visits();
  void                                      reorder();
};


TriangleReorderImpl::TriangleReorderImpl(Triangle* _tris, uint _n_tris, uint _n_vertices, uint _cache_size)
: tris_(_tris), n_tris_(_n_tris), n_vertices_(_n_vertices), 
  conn_(new Connectivity()), order_fpph_(*conn_), vertex_cache_(new VertexCache(_cache_size))
{}

TriangleReorderImpl::~TriangleReorderImpl()
{ 
  delete vertex_cache_; 
  order_fpph_.remove_property();
  delete conn_;
}

inline void TriangleReorderImpl::visit_vertex(int _vertex_index)
{
  assert(_vertex_index < (int)n_vertices_);
  if (!vertex_cache_->is_in(_vertex_index))
  {//not in the cache - has to be computed again
    vertex_cache_->insert(_vertex_index);
    n_visits_[_vertex_index]++;
  }
}

inline void TriangleReorderImpl::visit_face(int _face_index)
{
  visit_vertex(tris_[_face_index].v0);    
  visit_vertex(tris_[_face_index].v1);    
  visit_vertex(tris_[_face_index].v2);    
}

uint TriangleReorderImpl::calc_visits()
{
  vertex_cache_->clear();
  n_visits_.assign(n_vertices_, 0);  
  for (uint i = 0; i < n_tris_; ++i)
  {
    visit_face(i);
  }
  uint n_total_visits = std::accumulate(n_visits_.begin(), n_visits_.end(), 0);
  printf("Model consists of %d vertices which are visited (= uncached computation) %d times.\n"
         "Average %f visits per vertex. Rendering %d triangles. Cache size %d.\n", 
          n_vertices_, n_total_visits, (double)n_total_visits/(double)n_vertices_, 
          n_tris_, vertex_cache_->size());  
  return n_total_visits;
}


inline void TriangleReorderImpl::visit_vertex_ext(int _vertex_index)
{
  assert(_vertex_index < (int)n_vertices_);
  if (!vertex_cache_->is_in(_vertex_index))
  {//not in the cache - has to be computed again
    n_visits_[_vertex_index]++;//TODO: should we use this information ??
    int removed_index = vertex_cache_->insert(_vertex_index);
    if (removed_index >= 0)
    {
      for (Connectivity::ConstVertexFaceIter cvf_it = conn_->cvf_iter(VertexHandle(removed_index)); 
           cvf_it ; ++cvf_it)
      {
        if (order_fpph_[cvf_it] < 0 && calc_face_score(cvf_it) == 0)
        {//still not ordered - remove it from candidates if there are no other cached vertices referring to it
          assert(cand_fhs_.find(cvf_it) != cand_fhs_.end());
          cand_fhs_.erase(cvf_it);
        }
      }
    }
    for (Connectivity::ConstVertexFaceIter cvf_it = conn_->cvf_iter(VertexHandle(_vertex_index)); 
          cvf_it ; ++cvf_it)
    {
      if (order_fpph_[cvf_it] < 0)
      {//still not ordered - so insert it in the candidates set
        cand_fhs_.insert(cvf_it);
      }
    }
  }
}

inline void TriangleReorderImpl::visit_face_ext(int _face_index)
{
  visit_vertex_ext(tris_[_face_index].v0);    
  visit_vertex_ext(tris_[_face_index].v1);    
  visit_vertex_ext(tris_[_face_index].v2);    
}

inline void TriangleReorderImpl::order_face(FaceHandle _fh)
{
  assert(order_fpph_[_fh] < 0);
  order_fpph_[_fh] = next_id_++;
  cand_fhs_.erase(_fh);//remove this face from the candidates
  visit_face_ext(_fh.idx());
}

void TriangleReorderImpl::build_connectivity()
{
  conn_->clear();
  conn_->reserve(n_vertices_, 2*n_vertices_, n_tris_);
  conn_->resize(n_vertices_, 0, 0);
  
  for (uint i = 0; i < n_tris_; ++i)
  {
    FaceHandle fh = conn_->add_face(VertexHandle(tris_[i].v0), 
                                    VertexHandle(tris_[i].v1), VertexHandle(tris_[i].v2));
    order_fpph_[fh] = -1;
  }
  next_id_ = 0;
}

namespace
{
struct VertexDescr
{
  VertexHandle                              vh_;
  int                                       val_;
  
  VertexDescr(VertexHandle _vh, int _val)
  : vh_(_vh), val_(_val)
  {}
  
  inline bool operator < (const VertexDescr& _other) const
  { return val_ > _other.val_; }
};

typedef std::vector<VertexDescr>            VertexDescrArray;
}

void TriangleReorderImpl::build_avail_vhs()
{
  VertexDescrArray vertex_descr;
  vertex_descr.reserve(conn_->n_vertices());
  for (Connectivity::VertexIter v_it = conn_->vertices_begin(); v_it != conn_->vertices_end(); ++v_it)
  {
    vertex_descr.push_back(VertexDescr(v_it, conn_->valence(v_it)));
  }
  std::sort(vertex_descr.begin(), vertex_descr.end());
  avail_vhs_.clear();
  avail_vhs_.reserve(conn_->n_vertices());
  for (VertexDescrArray::const_iterator vd_it = vertex_descr.begin(); vd_it != vertex_descr.end(); ++vd_it)
  {
    avail_vhs_.push_back(vd_it->vh_);
  }
}

inline uint TriangleReorderImpl::calc_face_score(FaceHandle _fh) const
{
  assert(order_fpph_[_fh] == -1);//should be unprocessed
  int f_idx = _fh.idx();
  uint score = vertex_cache_->is_in(tris_[f_idx].v0) + vertex_cache_->is_in(tris_[f_idx].v1) +
               vertex_cache_->is_in(tris_[f_idx].v2);
  return score;
}

inline double TriangleReorderImpl::calc_face_score_ext(FaceHandle _fh) const
{ 
  assert(order_fpph_[_fh] == -1);//should be unprocessed
  int f_idx = _fh.idx();
  bool v0_in = vertex_cache_->is_in(tris_[f_idx].v0);
  bool v1_in = vertex_cache_->is_in(tris_[f_idx].v1);
  bool v2_in = vertex_cache_->is_in(tris_[f_idx].v2);

  double score = (uint)v0_in + (uint)v1_in + (uint)v2_in;


  // find number of un-visited neighbors
  for (Connectivity::ConstFaceFaceIter cff_it=conn_->cff_iter(_fh); cff_it; ++cff_it)
  {
    if (order_fpph_[cff_it] >= 0)
    {
      score += 0.01;
    }
  }

  return score;
}


FaceHandle TriangleReorderImpl::find_best_adjacent_face(VertexHandle _vh) const
{
  double max_score = -1;
  FaceHandle fh;
  for (Connectivity::ConstVertexFaceIter cvf_it = conn_->cvf_iter(_vh); cvf_it; ++cvf_it)
  {
    if (order_fpph_[cvf_it] < 0)
    {
      double score = calc_face_score_ext(cvf_it);
      if (score > max_score)
      {
        max_score = score;
        fh = cvf_it;
      }
    }
  }
  return fh;
}
  
FaceHandle TriangleReorderImpl::find_next_face()
{
//   for (int i = vertex_cache_->size() - 1; i >= 0; --i)
/*  for (uint i = 0, n = vertex_cache_->size(); i < n; ++i)
  {
    int v_id = (*vertex_cache_)[i];
    if (v_id >= 0)
    {//this cache entry is filled
      assert(v_id < (int)conn_->n_vertices());
      FaceHandle fh = find_best_adjacent_face(VertexHandle(v_id));
      if (fh.is_valid())
      {
        double score_fh = calc_face_score(fh);
        if (max_score < score_fh)
        {
          max_score = score_fh;
          best_fh = fh;
        }
      }
      if (fh.is_valid())
      {
        return fh;
      }
    }
  }*/
  double best_score = -1;
  FaceHandle best_fh;
  for (FaceHandleSet::const_iterator fh_it = cand_fhs_.begin(); fh_it != cand_fhs_.end(); ++fh_it)
  {
    double ext_score = calc_face_score_ext(*fh_it);
//     if (ext_score == 3.0)
//     {//we found a completely "free" face - exit here
//       return *fh_it;
//     }
    if (ext_score > best_score)
    {
      best_fh = *fh_it;
      best_score = ext_score;
    }
  }
  if (best_fh.is_valid())
  {
    return best_fh;
  }
  

  //no vertex which had unprocessed adjacent faces was found in the cache - so check the global vertices heap
  while (!avail_vhs_.empty())
  {
    VertexHandle vh = avail_vhs_.back();
    FaceHandle fh = find_best_adjacent_face(vh);
    if (fh.is_valid())
    {
      return fh;
    }
    else 
    {//this vertex does not have unprocessed adjacent faces - so remove it from the heap
      avail_vhs_.pop_back();
    }
  }
  assert(next_id_ == conn_->n_faces());


  return FaceHandle();//everything is processed 
}

void TriangleReorderImpl::reorder()
{
  if  (n_tris_ == 0 || n_vertices_ == 0 || vertex_cache_->size() == 0)
  {
    return;
  }
  build_connectivity();
  build_avail_vhs();
  vertex_cache_->clear();
  n_visits_.assign(n_vertices_, 0);  

  for (FaceHandle next_fh = find_next_face(); next_fh.is_valid(); next_fh = find_next_face())
  {
    order_face(next_fh);
  }
  reord_tris_.resize(conn_->n_faces());
  for (Connectivity::ConstFaceIter f_it = conn_->faces_begin(); f_it != conn_->faces_end(); ++f_it)
  {
    int curr_f_id = f_it.handle().idx();
    int new_f_id = order_fpph_[f_it];
    assert(0 <= new_f_id && new_f_id < (int)reord_tris_.size());
    reord_tris_[new_f_id] = tris_[curr_f_id];
  }
  memcpy(tris_, &reord_tris_[0], n_tris_*sizeof(Triangle));
}

/// TrianglReorder implementation

TriangleReorder::TriangleReorder(Triangle* _tris, uint _n_tris, uint _n_vertices, uint _cache_size)
: impl_(new TriangleReorderImpl(_tris, _n_tris, _n_vertices, _cache_size))
{}

TriangleReorder::~TriangleReorder()
{ delete impl_; }

uint TriangleReorder::calc_visits()
{ return impl_->calc_visits(); }

void TriangleReorder::reorder()
{ impl_->reorder(); }

}//namespace tr
