//=============================================================================
//
//  CLASS GLTrackball - IMPLEMENTATION
//
//=============================================================================

//== INCLUDES =================================================================

#ifdef WIN32
	#include <windows.h>
#define _USE_MATH_DEFINES 
#endif

#include <math.h>
#include <GL/gl.h>
#include <GL/glut.h>

#include "defs.h"
#include "GLTrackball.hh"


//== NAMESPACES ===============================================================


namespace ACG {


//== IMPLEMENTATION ========================================================== 


GLTrackball::GLTrackball() 
: center_(0,0,0) 
{
  for (int i=0; i<10; ++i)
    button_down_[i] = false;
}


//-----------------------------------------------------------------------------


void 
GLTrackball::mouse_press(int button, int x, int y)
{
  last_point_2D_ = Vec2i(x, y);
  last_point_ok_ = map_to_sphere(last_point_2D_, last_point_3D_);

  button_down_[button] = true;
}


//-----------------------------------------------------------------------------


void 
GLTrackball::mouse_release(int button, int x, int y)
{
  last_point_ok_ = false;
  button_down_[button] = false;

  // GLUT: button 3 or 4 -> mouse wheel clicked
  uint w = glutGet(GLUT_WINDOW_WIDTH);
  uint h = glutGet(GLUT_WINDOW_HEIGHT);
  if (button == 3)       
    zoom(0, (int)(last_point_2D_[1] - 0.05*w));
  else if (button == 4)
    zoom(0, (int)(last_point_2D_[1] + 0.05*h));
}


//-----------------------------------------------------------------------------


void 
GLTrackball::mouse_move(int x, int y)
{
  if (button_down_[0] && button_down_[1])
    action_ = ZOOM;
  else if (button_down_[0])
    action_ = ROTATION;
  else if (button_down_[1])
    action_ = TRANSLATION;

  switch (action_)
  {
    case ROTATION:    rotation(x, y);     break;
    case TRANSLATION: translation(x, y);  break;
    case ZOOM:        zoom(x, y);         break;
  }

  last_point_2D_ = Vec2i(x, y);
  last_point_ok_ = map_to_sphere(last_point_2D_, last_point_3D_);
}


//-----------------------------------------------------------------------------
struct GLMatrix
{
  typedef double Scalar;
  Scalar m[16];
  
  inline Scalar& operator ()(int i, int j) 
  {
      CHECK( 0 <= i && i <= 3);
      CHECK( 0 <= j && j <= 3);
      return m[(j<<2) + i];//columns linear in memory
  }
  
  inline Scalar  operator ()(int i, int j) const
  {
    CHECK( 0 <= i && i <= 3);
    CHECK( 0 <= j && j <= 3);
    return m[(j<<2) + i];//columns linear in memory
  }

  template <class Vec3>
  Vec3 operator * (const Vec3& a) const
  {
    Vec3 b;
  #define M(i,j) ((*this)(i,j))
  #define RES_I(i)  M(i,0)*a[0] + M(i,1)*a[1] + M(i,2)*a[2] + M(i,3);  
    b[0] = RES_I(0);
    b[1] = RES_I(1);
    b[2] = RES_I(2);
    Scalar w = RES_I(3);
    b /= w;
  #undef RES_I
  #undef M
    return b;
  }
};



void 
GLTrackball::rotation(int x, int y)
{
  if (last_point_ok_) 
  {
    Vec2i  new_point_2D_;
    Vec3f  new_point_3D_;
    bool   new_point_ok_;

    new_point_2D_ = Vec2i(x, y);
    new_point_ok_ = map_to_sphere(new_point_2D_, new_point_3D_);
    
    if (new_point_ok_)
    {
      Vec3f axis; 
      axis.set_cross(last_point_3D_, new_point_3D_);
      float cos_angle = last_point_3D_*new_point_3D_;

      if (fabs(cos_angle) < 1.0) 
      {
        float angle = 2.0*acos(cos_angle) * 180.0 / M_PI;
        GLMatrix modelview;
        OGL_CHECK(::glGetDoublev(GL_MODELVIEW_MATRIX, modelview.m));
        
        OGL_CHECK(::glMatrixMode(GL_MODELVIEW));
        OGL_CHECK(::glLoadIdentity());
        
        Vec3f t = modelview*center_;
        OGL_CHECK(::glTranslatef(t[0],  t[1],  t[2]));
        OGL_CHECK(::glRotatef(angle, axis[0], axis[1], axis[2]));
        OGL_CHECK(::glTranslatef(-t[0], -t[1], -t[2]));
        OGL_CHECK(::glMultMatrixd(modelview.m));
      }
    }
  }
}


//-----------------------------------------------------------------------------


void 
GLTrackball::translation(int x, int y)
{
/*  float dx = x - last_point_2D_[0];
  float dy = y - last_point_2D_[1];

  float z      = glstate_.modelview().transform_point(center_)[2];
  float w      = glutGet(GLUT_WINDOW_WIDTH);
  float h      = glutGet(GLUT_WINDOW_HEIGHT);
  float fovy   = glstate_.fovy();
  float nearpl = glstate_.near_plane();

  float aspect = w / h;
  float top    = tan(fovy/2.0f*M_PI/180.f) * nearpl;
  float right  = aspect*top;

  glstate_.translate(-2.0*dx/w*right/nearpl*z, 
                     2.0*dy/h*top/nearpl*z, 
                     0.0,
                     MULT_FROM_LEFT);
  */
}


//-----------------------------------------------------------------------------


void 
GLTrackball::zoom(int x, int y)
{
/*
  float dy = y - last_point_2D_[1];
  float z  = glstate_.modelview().transform_point(center_)[2];
  float h  = glstate_.viewport_height();

  glstate_.translate(0.0,
                     0.0,
                     -2.0*dy/h*z, 
                     MULT_FROM_LEFT);
  */
}


//-----------------------------------------------------------------------------


bool
GLTrackball::map_to_sphere(const Vec2i& _point, Vec3f& _result)
{
  float width  = glutGet(GLUT_WINDOW_WIDTH);
  float height = glutGet(GLUT_WINDOW_HEIGHT);
  
  if ( (_point[0] >= 0) && (_point[0] <= width) &&
       (_point[1] >= 0) && (_point[1] <= height) ) 
  {
    double x            = (_point[0] - 0.5*width)  / width;
    double y            = (0.5*height - _point[1]) / height;
    double sinx         = sin(M_PI * x * 0.5);
    double siny         = sin(M_PI * y * 0.5);
    double sinx2siny2   = sinx * sinx + siny * siny;
    
    _result[0] = sinx;
    _result[1] = siny;
    _result[2] = sinx2siny2 < 1.0 ? sqrt(1.0 - sinx2siny2) : 0.0;
    
    return true;
  }
  else return false;
}


//=============================================================================
} // namespace ACG
//=============================================================================
