/******************************************************************************
                       C++ Implementation: TSMVisualARB
                            -------------------
    description       :
    begin             : Wed Mar 23 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#include "TSMVisualARB.hh"

#include <stdio.h>
#include <stdlib.h>
#include <sstream>

#include <GL/glew.h>

#include "vbo.h"
#include "TwoScaleModel.h"

#include "half.hh"



namespace
{

IVisualization* create_ARB_vis(const TwoScaleModel& _tsm, int argc, char** argv)
{ 
  for (int i=1; i < argc; ++i)
  {
    if (strcmp(argv[i], "-vp") == 0 && (i + 1) < argc)
    {//set "vertex program for atributes" file name
      return new TSMVisualARB(_tsm, argv[i+1]);
    }
  }  
  printf("Failed to find the path to the ARB vertex program sources.\n" 
         "Usage: -vp path .\n");
  return NULL;
}

struct RegisterARBVisCreator
{
  RegisterARBVisCreator()
  { VisualizationFactory::connect('6', create_ARB_vis); }
};

RegisterARBVisCreator register_vp_vis_creator__;

}

TSMVisualARB::TSMVisualARB(const TwoScaleModel& _tsm, const char* _vp_path)
  : tsm_(_tsm), vp_path_(_vp_path),
    light_map_(GL_TEXTURE0_ARB)
{}


TSMVisualARB::~TSMVisualARB()
{
  deactivate();
  unload_vertex_program();
}

void TSMVisualARB::unload_vertex_program()
{
  vertex_program_.unbind();
  glDisable(GL_VERTEX_PROGRAM_ARB);
}



bool TSMVisualARB::init()
{
  uint n_fs_vertices = tsm_.fs_vertices_.size();
  std::vector<Vec3> foot_pt(n_fs_vertices), foot_du(n_fs_vertices), foot_dv(n_fs_vertices);
  TwoScaleModel::CVCoeffArray cv_coeff;
  uint n_modified_cvs = tsm_.calc_geometry_split(&foot_pt.front(), 
						 &foot_du.front(), 
						 &foot_dv.front(), 
						 cv_coeff);
  

  // collect modified control vertices
  modified_cvs_.clear(); 
  modified_cvs_.reserve(n_modified_cvs);
  for (uint i = 0, n = cv_coeff.size(); i < n; ++i)
    if (!cv_coeff[i].empty())
      modified_cvs_.push_back(i);


  union { unsigned short int s[2]; float f; } d32;
#define add_float32(F)  d32.f=F; data.push_back(d32.s[0]); data.push_back(d32.s[1]);
#define add_float16(F)  data.push_back( half(F).bits() );


  // setup data for VBO
  std::vector<short> data; 
  data.reserve(n_fs_vertices * (16 + 3*modified_cvs_.size()));

  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    add_float32( tsm_.fs_vertices_[i].t0_displ_n_  );
    add_float32( tsm_.fs_vertices_[i].t0_displ_du_ );
    add_float32( tsm_.fs_vertices_[i].t0_displ_dv_ );
    add_float32( tsm_.fs_vertices_[i].pt_displ_    );

    add_float32( foot_pt[i][0] );
    add_float32( foot_pt[i][1] );
    add_float32( foot_pt[i][2] );
    add_float32( tsm_.fs_vertices_[i].t1_displ_n_ );
    
    add_float32( foot_du[i][0] );
    add_float32( foot_du[i][1] );
    add_float32( foot_du[i][2] );
    add_float32( tsm_.fs_vertices_[i].t1_displ_du_ );
    
    add_float32( foot_dv[i][0] );
    add_float32( foot_dv[i][1] );
    add_float32( foot_dv[i][2] );
    add_float32( tsm_.fs_vertices_[i].t1_displ_dv_ );

    for (uint j=0, n=modified_cvs_.size(); j<n; ++j)
    {
      uint cv_id = modified_cvs_[j]; 
      CHECK(!cv_coeff[cv_id].empty());
      
      add_float16( cv_coeff[cv_id][i][0] );
      add_float16( cv_coeff[cv_id][i][1] );
      add_float16( cv_coeff[cv_id][i][2] );
      add_float16( 0.0 );
    }
  }


  // geometry buffer
  uint mem = data.size()*sizeof(data[0]);
  uint stride = mem / n_fs_vertices;
  vbo_.bind();
  vbo_.upload( mem, &data[0], GL_STATIC_DRAW_ARB );


  // array pointers
  OGL_CHECK(::glVertexAttribPointerARB( 0, 4, GL_FLOAT, false, stride, vbo_.offset( 0*4) ));
  OGL_CHECK(::glVertexAttribPointerARB( 1, 4, GL_FLOAT, false, stride, vbo_.offset( 4*4) ));
  OGL_CHECK(::glVertexAttribPointerARB( 2, 4, GL_FLOAT, false, stride, vbo_.offset( 8*4) ));
  OGL_CHECK(::glVertexAttribPointerARB( 3, 4, GL_FLOAT, false, stride, vbo_.offset(12*4) ));
  for (uint j=0, n=modified_cvs_.size(); j<n; ++j)
    OGL_CHECK(::glVertexAttribPointerARB( 4+j, 4, GL_HALF_FLOAT_NV, false, stride, vbo_.offset(16*4 + 4*j*2) ));


  // index buffer
  n_indices_   = tsm_.fs_tris_.size()*3;
  ibo_.bind();
  ibo_.upload( n_indices_*sizeof(int), &tsm_.fs_tris_[0], GL_STATIC_DRAW_ARB );
  

  // update light map
  light_map_.update();
  light_map_.bind();


  // setup program
  std::ostringstream  os;
#ifdef WIN32
  os << vp_path_ << "\\vp_deform" << modified_cvs_.size() << ".h";
#else
  os << vp_path_ << "/vp_deform" << modified_cvs_.size() << ".h";
#endif
  if (!load_vertex_program(os.str().c_str())) return false;
  activate();


  return true;
}



bool TSMVisualARB::load_vertex_program(const char* _filename)
{
  unload_vertex_program();

  if (!vertex_program_.load_file(_filename))
  {
    printf("ERROR: Loading of vertex shader source file: %s failed!\n", _filename);
    return false;
  }

  return true;
}


void TSMVisualARB::activate()
{
  vertex_program_.bind();
  glEnable(GL_VERTEX_PROGRAM_ARB);

  vbo_.bind();
  glEnableVertexAttribArrayARB(0);
  glEnableVertexAttribArrayARB(1);
  glEnableVertexAttribArrayARB(2);
  glEnableVertexAttribArrayARB(3);
  for (uint j=0, n=modified_cvs_.size(); j<n; ++j)
    glEnableVertexAttribArrayARB(4+j);

  light_map_.bind();
  light_map_.enable();
}


void TSMVisualARB::deactivate()
{
  vertex_program_.unbind();
  glDisable(GL_VERTEX_PROGRAM_ARB);

  vbo_.unbind();
  vbo_.del();

  ibo_.unbind();
  ibo_.del();

  for (uint j=0; j<16; ++j)
  {
    glDisableVertexAttribArrayARB(j);
    glVertexAttribPointerARB( j, 4, GL_FLOAT, false, 0, NULL );
  }

  light_map_.disable();
}


void TSMVisualARB::draw(ModeEnum _mode)
{
  /* the control fsv_blocks positions which we upload to the GPU here
     are the sole information which is transfered to the GPU at every
     frame */
  for (uint i = 0, n = modified_cvs_.size(); i < n; ++i)
  {
    const Vec3& cv = tsm_.cm_vertices_[modified_cvs_[i]];
    OGL_CHECK(::glProgramEnvParameter4fARB(GL_VERTEX_PROGRAM_ARB, i, 
					   cv.x, cv.y, cv.z, 0.0));
  }


  //draw now
  if (_mode == MODE_POINTS)
  {
    OGL_CHECK(::glDrawArrays(GL_POINTS, 0, tsm_.fs_vertices_.size()));
  }
  else if (tsm_.prim_groups_ != NULL)
  {
    static uint ogl_pt[PT_FAN+1] = {GL_TRIANGLES, GL_TRIANGLE_STRIP, GL_TRIANGLE_FAN};
    for (uint i = 0; i < tsm_.n_prim_groups_; ++i)
    {
      const PrimitiveGroup& pg = tsm_.prim_groups_[i];
      OGL_CHECK(::glDrawElements(ogl_pt[pg.type], pg.numIndices, GL_UNSIGNED_INT, pg.indices));
    }
  }
  else
  {
    OGL_CHECK(::glDrawElements(GL_TRIANGLES, n_indices_, GL_UNSIGNED_INT, ibo_.offset(0)));
  }
}


void TSMVisualARB::dump()
{ 
}

