/******************************************************************************
                       C++ Implementation: TSMVisualVT
                            -------------------
    description       :
    begin             : Wed Apr 20 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#include "TSMVisualVT.h"

#ifdef WIN32
  #include <windows.h>
#endif

#include "vbo.h"
#include "TwoScaleModel.h"
#include "VertexProgram.h"


namespace
{

IVisualization* create_VT_vis(const TwoScaleModel& _tsm, int argc, char** argv)
{ 
  for (int i=1; i < argc; ++i)
  {
    if (strcmp(argv[i], "-vst") == 0 && (i + 1) < argc)
    {//set "vertex program for textures" file name
      return new TSMVisualVT(_tsm, argv[i+1]);
    }
  }  
  return NULL;
}

struct RegisterVTVisCreator
{
  RegisterVTVisCreator()
  { VisualizationFactory::connect('3', create_VT_vis); }
};

RegisterVTVisCreator register_vt_vis_creator__;

}

template <class T>
uint new_tex(const std::vector<T>& _data, int& _w, int& _h)
{
//   const uint tex_comp_size__ = 4*sizeof(float);//GL_RGBA_FLOAT32_ATI textures
  CHECK_STATIC(sizeof(T) == 4*sizeof(float));//only alligned 4 floats data allowed
  int max_tex_size; 
  glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_tex_size);
  _w = max_tex_size;
  _h = 1 + _data.size()/_w;
  if (_h > max_tex_size)
  {//data size too large - it is not possible to create so large 2D texture
    return 0;
  }
  uint tex_buf_size = (uint)_w*_h;
  CHECK(tex_buf_size >= _data.size());
  
  T* tex_buf = new T[tex_buf_size];
  for (uint i = 0, n = _data.size(); i < n; ++i)
  {//copy the data to the new buffer
    tex_buf[i] = _data[i];
  }

  uint tex_id;
  OGL_CHECK(::glGenTextures(1, &tex_id));
  OGL_CHECK(::glBindTexture(GL_TEXTURE_2D, tex_id));
  OGL_CHECK(::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST));
  OGL_CHECK(::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST));
  OGL_CHECK(::glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, _w, _h, 0, GL_RGBA, GL_FLOAT, tex_buf));
                           
  /* TODO: use sizeof(T)/sizeof(float) GL_LUMINANCE_FLOAT32 TEXTURE instead for control vertices as
      the fourth component will be not used anyway? */
  delete [] tex_buf;

  return tex_id;
}

inline void delete_tex(uint _handle)
{ OGL_CHECK(::glDeleteTextures(1, &_handle)); }

inline bool tex_is_resident(uint _handle)
{
  GLboolean res;
  OGL_CHECK(res = ::glAreTexturesResident(1, &_handle, &res));
  return res == GL_TRUE;
}

TSMVisualVT::TSMVisualVT(const TwoScaleModel& _tsm, const std::string& _vs_file_name)
: tsm_(_tsm), vs_file_name_(_vs_file_name), vertex_program_(NULL)
{}

TSMVisualVT::~TSMVisualVT()
{
  deactivate();
  unload_vertex_program();
  //clrer constant data
  delete_VBO(vbo_vertices_);
  delete_VBO(vbo_normals_);//test data
  delete_VBO(vbo_tex_coords_);
  delete_VBO(vbo_indices_);

  delete_tex(tex_coeffs_);
  delete_tex(tex_cvs_);
//  delete_tex(tex_vertices_);
//  delete_tex(tex_normals_);

}

void TSMVisualVT::unload_vertex_program()
{
  if (vertex_program_ != NULL)
  {
    vertex_program_->stop();
    delete vertex_program_;      
    vertex_program_ = NULL;
  }
}

bool TSMVisualVT::init()
{
  if (vs_file_name_.empty())
  {
    printf("ERROR: No vertex shader file name specified.\n");
    return false;
  }
  
  unload_vertex_program();
  
  vertex_program_ = new VertexProgram();
  
  printf("Loading vertex shader source: %s ...\n", vs_file_name_.c_str());
  if (!vertex_program_->load_source(vs_file_name_.c_str()))
  {
    printf("ERROR: Loading of vertex shader source file: %s failed!\n", vs_file_name_.c_str());
    delete vertex_program_;
    vertex_program_ = NULL;
    return false;
  }
  if (!vertex_program_->compile_and_link())
  {
    printf("ERROR: Compilation/linking of the vertex program failed!" 
           "Dump the source for more information.\n");
    return false;
  }

  return init_data();
}

namespace
{

struct CVC
{
  float cv_id_;
  float f_, du_, dv_;

  CVC()  {}

  inline CVC(const ControlVertexCoeff& _cvc)
  : cv_id_((float)_cvc.cv_id_), f_(_cvc.f_), du_(_cvc.du_), dv_(_cvc.dv_)
  {}
};

}

bool TSMVisualVT::init_tex_coeffs()
{

  uint n_fs_vertices = tsm_.fs_vertices_.size();

  std::vector<CVC> coeffs;
  std::vector<std::pair<uint, uint> > coeff_indices(n_fs_vertices);

  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    coeff_indices[i].first = coeffs.size();//first coefficient's index
    
    const FineScaleVertex& fsv = tsm_.fs_vertices_[i];
    for (uint j = 0, m = fsv.cv_coeffs_.size(); j < m; ++j)
    {
      coeffs.push_back(CVC(fsv.cv_coeffs_[j]));
    }
    coeff_indices[i].second = coeffs.size();//past-the-end coefficient's index
  }
  vbo_tex_coords_ = new_VBO(coeff_indices);
  tex_coeffs_ = new_tex(coeffs, tex_coeffs_w_, tex_coeffs_h_);

  return tex_coeffs_ != 0;
}

bool TSMVisualVT::init_tex_cvs()
{
  std::vector<Vec3Wt> cvs4(tsm_.cm_vertices_.size());
  for (uint i = 0; i < cvs4.size(); ++i)
  {
    cvs4[i].v = tsm_.cm_vertices_[i];
    cvs4[i].w = 1;
  }
  tex_cvs_ = new_tex(cvs4, tex_cvs_w_, tex_cvs_h_);
  
  return tex_cvs_ != 0;
}

bool TSMVisualVT::init_data()
{
  //displacements
  uint n_fs_vertices = tsm_.fs_vertices_.size();
  std::vector<Vec3Wt> displ(n_fs_vertices);
  tsm_.get_displacements(&displ.front());
  vbo_vertices_ = new_VBO(displ);    
  
/*//test data
  uint n_fs_vertices = tsm_.fs_vertices_.size();
  std::vector<Vec3> vertices(n_fs_vertices), normals(n_fs_vertices);
  tsm_.calc_geometry(&vertices.front(), &normals.front());
  
  vbo_vertices_ = new_VBO(vertices);
  vbo_normals_ = new_VBO(normals);
*/  
  
  //indices
  n_indices_ = tsm_.fs_tris_.size()*3;
  vbo_indices_ = new_VBO(tsm_.fs_tris_, GL_ELEMENT_ARRAY_BUFFER_ARB);

  if (!init_tex_coeffs())
  {
    printf("Failed to create the displacements texture. Dimensions: (%d, %d).\n",
           tex_coeffs_w_, tex_coeffs_h_);
    return false;
  }
  if (!init_tex_cvs())
  {
    printf("Failed to create the control vertices texture. Dimensions: (%d, %d).\n",
           tex_cvs_w_, tex_cvs_h_);
    return false;
  }  
  tex_cvs_buf_.resize(tex_cvs_w_*tex_cvs_h_);
  activate();
  return true;
}

void TSMVisualVT::activate()
{
  vertex_program_->start();  
  
  //setup coefficients' texture bindings with the vertex shader
  loc_tex_coeffs_ = vertex_program_->query_uniform_location("tex_coeffs");
  loc_tex_coeffs_w_ = vertex_program_->query_uniform_location("tex_coeffs_w");
  loc_tex_coeffs_h_ = vertex_program_->query_uniform_location("tex_coeffs_h");

  OGL_CHECK(::glUniform1iARB(loc_tex_coeffs_, 0));
  OGL_CHECK(::glUniform1fARB(loc_tex_coeffs_w_, (float)tex_coeffs_w_));
  OGL_CHECK(::glUniform1fARB(loc_tex_coeffs_h_, (float)tex_coeffs_h_));

  //setup cvs' texture bindings with the vertex shader
  loc_tex_cvs_ = vertex_program_->query_uniform_location("tex_cvs");
  loc_tex_cvs_w_ = vertex_program_->query_uniform_location("tex_cvs_w");
  loc_tex_cvs_h_ = vertex_program_->query_uniform_location("tex_cvs_h");

  OGL_CHECK(::glUniform1iARB(loc_tex_cvs_, 1));
  OGL_CHECK(::glUniform1fARB(loc_tex_cvs_w_, (float)tex_cvs_w_));
  OGL_CHECK(::glUniform1fARB(loc_tex_cvs_h_, (float)tex_cvs_h_));

  OGL_CHECK(::glEnableClientState(GL_VERTEX_ARRAY));
//  OGL_CHECK(::glEnableClientState(GL_NORMAL_ARRAY));// test data
  OGL_CHECK(::glEnableClientState(GL_TEXTURE_COORD_ARRAY));
}


void TSMVisualVT::deactivate()
{
  if (vertex_program_ != NULL)
  {
    vertex_program_->stop();  
  }
  
  OGL_CHECK(::glDisableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glDisableClientState(GL_TEXTURE_COORD_ARRAY));
//  OGL_CHECK(::glDisableClientState(GL_NORMAL_ARRAY));// test data
  
  OGL_CHECK(::glActiveTexture(GL_TEXTURE0));
  OGL_CHECK(::glBindTexture(GL_TEXTURE_2D, 0));
  OGL_CHECK(::glActiveTexture(GL_TEXTURE1));
  OGL_CHECK(::glBindTexture(GL_TEXTURE_2D, 0));
}


void TSMVisualVT::draw(ModeEnum _mode)
{
  OGL_CHECK(::glActiveTexture(GL_TEXTURE0));
  OGL_CHECK(::glBindTexture(GL_TEXTURE_2D, tex_coeffs_));
  
  OGL_CHECK(::glActiveTexture(GL_TEXTURE1));
  OGL_CHECK(::glBindTexture(GL_TEXTURE_2D, tex_cvs_));
  //update cvs texture
  
  CHECK(tex_cvs_buf_.size() == (uint)tex_cvs_w_*tex_cvs_h_);
  for (uint i = 0, n = tsm_.cm_vertices_.size(); i < n; ++i)
  {
    tex_cvs_buf_[i].v = tsm_.cm_vertices_[i];
  }
  OGL_CHECK(::glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, tex_cvs_w_, tex_cvs_h_, GL_RGBA, GL_FLOAT, 
                              &tex_cvs_buf_[0]));
/*  //test data
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vertices_));
  OGL_CHECK(::glVertexPointer(3, GL_FLOAT, 0, (char *) NULL));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_normals_));
  OGL_CHECK(::glNormalPointer(GL_FLOAT, 0, (char *) NULL));
*/
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vertices_));
  OGL_CHECK(::glVertexPointer(4, GL_FLOAT, 0, (char *) NULL));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_tex_coords_));
  OGL_CHECK(::glTexCoordPointer(2, GL_INT, 0, (char *) NULL));

  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer
  
  //upload control vertices' data with glTexSubImage
  if (_mode == GL_POINTS)
  {
    OGL_CHECK(::glDrawArrays(GL_POINTS, 0, tsm_.fs_vertices_.size()));
  }
  else
  {
    OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbo_indices_));
    OGL_CHECK(::glDrawElements(GL_TRIANGLES, n_indices_, GL_UNSIGNED_INT, (char *) NULL));
    OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0));//unbind indicies buffer
  }
}

void TSMVisualVT::dump()
{
/*  if (vertex_program_ == NULL)
  {
    printf("ERROR: vertex program not loaded.\n");
  }
  else
  {
    vertex_program_->dump_source(); 
  }
*/
  printf("Coefficients' texture is %sresident.\n", tex_is_resident(tex_coeffs_) ? "" : "NOT ");
  printf("Control vertices' texture is %sresident.\n", tex_is_resident(tex_coeffs_) ? "" : "NOT ");
}




  /* //texture coordinates
  std::vector<uint> v_idx(n_fs_vertices);
  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    v_idx[i] = i;
  }
  vbo_tex_coords_ = new_VBO(v_idx);

  std::vector<Vec3Wt> vertices4(n_fs_vertices), normals4(n_fs_vertices);
  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    vertices4[i].v = vertices[i];
    vertices4[i].w = 1;
    normals4[i].v = normals[i];
    normals4[i].w = 1;
  }
  tex_vertices_ = new_tex(vertices4, tex_w_, tex_h_);
  tex_normals_ = new_tex(normals4, tex_w_, tex_h_);
  if (tex_vertices_ == 0 || tex_normals_ == 0)
  {
    printf("Failed creating texture(s) with size (%d,%d).\n", tex_w_, tex_h_);
    return false;
  }
  */


/*
void TSMVisualVT::activate()
{
  vertex_program_->start();  
  
  tex_vertices_loc_ = vertex_program_->query_uniform_location("tex_vertices");
  tex_normals_loc_ = vertex_program_->query_uniform_location("tex_normals");
  tex_w_loc_ = vertex_program_->query_uniform_location("tex_w");
  tex_h_loc_ = vertex_program_->query_uniform_location("tex_h");
  
  OGL_CHECK(::glUniform1iARB(tex_vertices_loc_, 0));
  OGL_CHECK(::glUniform1iARB(tex_normals_loc_, 1));
  OGL_CHECK(::glUniform1fARB(tex_w_loc_, (float)tex_w_));
  OGL_CHECK(::glUniform1fARB(tex_h_loc_, (float)tex_h_));

  OGL_CHECK(::glActiveTexture(GL_TEXTURE0));
  OGL_CHECK(::glBindTexture(GL_TEXTURE_2D, tex_vertices_));
  
  OGL_CHECK(::glActiveTexture(GL_TEXTURE1));
  OGL_CHECK(::glBindTexture(GL_TEXTURE_2D, tex_normals_));
      
  OGL_CHECK(::glEnableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glEnableClientState(GL_NORMAL_ARRAY));
  OGL_CHECK(::glEnableClientState(GL_TEXTURE_COORD_ARRAY));
}
*/

