/******************************************************************************
                       C++ Implementation: TwoScaleModel
                            -------------------
    description       :
    begin             : Wed Mar 23 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#include "TwoScaleModel.h"
#include <stdlib.h>
#include "../TriangleReorder/TriangleReorder.h"
#include "../TriStripper/tri_stripper.h"

#define BUF_SIZE   2048

#define CHECK_FSCANF_RES(EXPR) { int res = EXPR; if (res == 0 || res == EOF) return false; }

TwoScaleModel::TwoScaleModel()
: prim_groups_(NULL), n_prim_groups_(0)
{}

TwoScaleModel::~TwoScaleModel()
{ clear_prim_groups(); }

bool TwoScaleModel::load(const char* _file_name)
{
  FILE* in_file = fopen(_file_name, "r");
  if (in_file == NULL)
  {
    return false;
  }
  // coarse scale 
  uint n_cm_vertices, n_cm_faces;
  CHECK_FSCANF_RES(fscanf(in_file, "%d %d", &n_cm_vertices, &n_cm_faces));
  cm_vertices_.resize(n_cm_vertices);

  for (uint i = 0; i < n_cm_vertices; ++i)
  {
    Vec3& v = cm_vertices_[i];
    CHECK_FSCANF_RES(fscanf(in_file, "%f %f %f", &v.x, &v.y, &v.z));
  }
  for (uint i = 0; i < n_cm_faces; ++i)//we don't need these
  {
    uint f_val;//valence of the control face
    CHECK_FSCANF_RES(fscanf(in_file, "%d ", &f_val));
    for (uint j = 0; j < f_val; ++j)
    {//skip control mesh vertex indices
      int tmp;
      CHECK_FSCANF_RES(fscanf(in_file, "%d ", &tmp));
    }
  }
  cv_modified_.resize(cm_vertices_.size(), false);
  
  // fine scale 
  uint n_fs_vertices, n_fs_tris;  
  CHECK_FSCANF_RES(fscanf(in_file, "%d %d", &n_fs_vertices, &n_fs_tris));
  
  fs_vertices_.resize(n_fs_vertices);
  
  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    FineScaleVertex& v = fs_vertices_[i];
    CHECK_FSCANF_RES(fscanf(in_file, "%f%*[^(]%*c%f %f %f%*c%*[^(]%*c%f %f %f%*c", 
                                     &v.pt_displ_, 
                                     &v.t0_displ_n_, &v.t0_displ_du_, &v.t0_displ_dv_,
                                     &v.t1_displ_n_, &v.t1_displ_du_, &v.t1_displ_dv_));
    uint n_coeffs;
    CHECK_FSCANF_RES(fscanf(in_file, "%d", &n_coeffs));  
    v.cv_coeffs_.resize(n_coeffs);
    for (uint j = 0; j < n_coeffs; ++j)
    {
      fscanf(in_file, "%*[^(]%*c%d %f %f %f%*c", &v.cv_coeffs_[j].cv_id_, &v.cv_coeffs_[j].f_, 
                                                 &v.cv_coeffs_[j].du_, &v.cv_coeffs_[j].dv_);
    }
  }
    
  fs_tris_.resize(n_fs_tris);
  for (uint i = 0; i < n_fs_tris; ++i)
  {
    CHECK_FSCANF_RES(fscanf(in_file, "%d %d %d", &fs_tris_[i].v0, &fs_tris_[i].v1, &fs_tris_[i].v2));
  }
  
  fclose(in_file);
  return true;
}
    
void TwoScaleModel::calc_geometry(Vec3* _vertices,  Vec3* _normals) const
{
  Vec3* vertex_it = _vertices;
  Vec3* normal_it = _normals;
  
  for (FineScaleVertexArray::const_iterator fsv_it = fs_vertices_.begin(), fsv_end = fs_vertices_.end();       
       fsv_it != fsv_end; ++fsv_it, ++vertex_it, ++normal_it)
  {//iterate all fine scale vertices 
    Vec3 foot_pt(0), foot_du(0), foot_dv(0);
    for (FineScaleVertex::ControlVertexCoeffArray::const_iterator cvc_it = fsv_it->cv_coeffs_.begin(); 
         cvc_it != fsv_it->cv_coeffs_.end(); ++cvc_it)
    {//compute the foot_*  as linear combination of the _cm vertices
      const Vec3& cv_pt = cm_vertices_[cvc_it->cv_id_];
      foot_pt.add_linear(cvc_it->f_, cv_pt);
      foot_du.add_linear(cvc_it->du_, cv_pt);
      foot_dv.add_linear(cvc_it->dv_, cv_pt);
    }
    Vec3 foot_sn;
    foot_sn.set_cross(foot_du, foot_dv);
    foot_sn.normalize();
    //displace foot_pt in the local frame toward the fine scale point
    foot_pt.add_linear(fsv_it->pt_displ_, foot_sn);
    //calc the normal as a linear combination of the local frame
    Vec3 t0, t1;
    t0.set_linear(fsv_it->t0_displ_n_ , foot_sn);
    t0.add_linear(fsv_it->t0_displ_du_, foot_du);
    t0.add_linear(fsv_it->t0_displ_dv_, foot_dv);
    
    t1.set_linear(fsv_it->t1_displ_n_ , foot_sn);
    t1.add_linear(fsv_it->t1_displ_du_, foot_du);
    t1.add_linear(fsv_it->t1_displ_dv_, foot_dv);
    
    *vertex_it = foot_pt; 
    normal_it->set_cross(t0, t1);
  }
}

void TwoScaleModel::get_displacements(Vec3Wt* _displ) const
{
  //TODO: remove this API entry
  Vec3Wt* displ_it = _displ;
  for (FineScaleVertexArray::const_iterator fsv_it = fs_vertices_.begin(), fsv_end = fs_vertices_.end();       
       fsv_it != fsv_end; ++fsv_it, ++displ_it)
  {//iterate all fine scale vertices
/*    displ_it->v = fsv_it->n_displ_;
    displ_it->w = fsv_it->pt_displ_;*/
    displ_it->v = Vec3(1, 0, 0);
    displ_it->w = 0;
  }
}

uint TwoScaleModel::calc_geometry_split(Vec3* _foot_pt, Vec3* _foot_du, Vec3* _foot_dv,
                                        CVCoeffArray& _cv_coeff) const
{
  uint n_modifed_cvs = 0;
  _cv_coeff.resize(cm_vertices_.size());
  for (uint i = 0, n = cm_vertices_.size(); i < n; ++i)
  {//create coefficients arrays only for the modified vertices
    if (cv_modified_[i])
    {//init the coefficients array with 0, which is correct for all vertices independent from cv_i
      _cv_coeff[i].assign(fs_vertices_.size(), Vec3(0));
      n_modifed_cvs++;
    }
    else
    {
      _cv_coeff[i].clear();
    }
  }
  Vec3* foot_pt_it = _foot_pt;
  Vec3* foot_du_it = _foot_du;
  Vec3* foot_dv_it = _foot_dv;
  for (FineScaleVertexArray::const_iterator fsv_it = fs_vertices_.begin(), fsv_end = fs_vertices_.end();       
       fsv_it != fsv_end; ++fsv_it, ++foot_pt_it, ++foot_du_it, ++foot_dv_it)
  {//iterate all fine scale vertices 
    Vec3 foot_pt(0), foot_du(0), foot_dv(0);
    uint fsv_id = fsv_it - fs_vertices_.begin();
    for (FineScaleVertex::ControlVertexCoeffArray::const_iterator cvc_it = fsv_it->cv_coeffs_.begin(); 
         cvc_it != fsv_it->cv_coeffs_.end(); ++cvc_it)
    {//compute the foot_*  as linear combination of the _cm vertices
      if (cv_modified_[cvc_it->cv_id_])
      {
        CHECK(_cv_coeff[cvc_it->cv_id_].size() == fs_vertices_.size());
        _cv_coeff[cvc_it->cv_id_][fsv_id] = Vec3(cvc_it->f_, cvc_it->du_, cvc_it->dv_);
      }
      else
      {//this control vertex is static so add its contribution to the precomputed foot_* arrays
        const Vec3& cv_pt = cm_vertices_[cvc_it->cv_id_];
        foot_pt.add_linear(cvc_it->f_, cv_pt);
        foot_du.add_linear(cvc_it->du_, cv_pt);
        foot_dv.add_linear(cvc_it->dv_, cv_pt);
      }
    }
    *foot_pt_it = foot_pt;
    *foot_du_it = foot_du;
    *foot_dv_it = foot_dv;
  }
  return n_modifed_cvs;
}

void TwoScaleModel::dump_stats() const
{
  uint n_cv_coeffs = 0;
  uint n_max_cv_coeffs = 0;
  uint n_zero_coeff = 0;
  for (FineScaleVertexArray::const_iterator fsv_it = fs_vertices_.begin(), fsv_end = fs_vertices_.end();       
       fsv_it != fsv_end; ++fsv_it)
  {
    n_cv_coeffs += fsv_it->cv_coeffs_.size();
    htf::set_max(n_max_cv_coeffs, fsv_it->cv_coeffs_.size());
    for (FineScaleVertex::ControlVertexCoeffArray::const_iterator cvc_it = fsv_it->cv_coeffs_.begin(); 
         cvc_it != fsv_it->cv_coeffs_.end(); ++cvc_it)
    {
      if (cvc_it->f_ == 0.0f)
      {
        n_zero_coeff++;
      }
    }
  }
  printf("Control vertices number : %d.\n", cm_vertices_.size());
  printf("Fine scale vertices/faces number : %d/%d.\n", fs_vertices_.size(), fs_tris_.size());
  printf("Total/average/max number of CV coefficients: %d, %f, %d.\n", 
                  n_cv_coeffs, (double)n_cv_coeffs/(double)fs_vertices_.size(), n_max_cv_coeffs);
  printf("Zero CV coefficients: %d, %f%%\n", n_zero_coeff, (double)n_zero_coeff/(double)n_cv_coeffs);
  double total_mem = (double)(n_cv_coeffs*sizeof(ControlVertexCoeff) + 2*sizeof(Vec3))/(1024.0*1024.0);
  printf("Memory required for fine scale description: %fMb.\n", total_mem);
}

void TwoScaleModel::optimize_NvTriStrip(bool _create_strips)
{
  clear_prim_groups();
  ::SetCacheSize(24);
  ::SetListsOnly(!_create_strips);
  ::SetStitchStrips(false);
  ::GenerateStrips((const uint*)&fs_tris_[0], fs_tris_.size()*3, &prim_groups_, &n_prim_groups_);

  
  if (n_prim_groups_ == 1 && prim_groups_[0].type == PT_LIST && prim_groups_[0].numIndices == fs_tris_.size()*3)
  {
    printf("NvTriStrip returned a single triangle list.\n");
    printf("Remapping indices & vertices.\n");
    PrimitiveGroup* remapped_pgs;
    ::RemapIndices(prim_groups_, n_prim_groups_, fs_vertices_.size(), &remapped_pgs);
    FineScaleVertexArray remapped_fs_vertices(fs_vertices_.size());
    
    for (uint i = 0; i < n_prim_groups_; ++i)
    {
      const PrimitiveGroup& orig_pg = prim_groups_[i];
      const PrimitiveGroup& remap_pg = remapped_pgs[i];
      for (uint j = 0; j < orig_pg.numIndices; ++j)
      {
        CHECK(orig_pg.indices[j] < fs_vertices_.size() && remap_pg.indices[j] < fs_vertices_.size());
        remapped_fs_vertices[remap_pg.indices[j]] = fs_vertices_[orig_pg.indices[j]];
      }
    }
    delete [] prim_groups_;
    prim_groups_ = remapped_pgs;
    fs_vertices_ = remapped_fs_vertices;    
    
    memcpy(&fs_tris_[0], prim_groups_[0].indices, fs_tris_.size()*3*sizeof(uint));
    clear_prim_groups();
  }
  else
  {
    printf("%d primitve groups returned from NvTriStrip.\n", n_prim_groups_);
  }
}

void TwoScaleModel::optimize_TriStripper()
{
  uint n_fs_tris = fs_tris_.size();
  triangle_stripper::indices indices(3*n_fs_tris);
  for (uint i = 0; i < n_fs_tris; ++i)
  {
    indices[3*i  ] = fs_tris_[i].v0;
    indices[3*i+1] = fs_tris_[i].v1;
    indices[3*i+2] = fs_tris_[i].v2;
  }
  
  triangle_stripper::primitive_vector  primitives;
  triangle_stripper::tri_stripper      stripper(indices);
  
  stripper.SetCacheSize(18);
  stripper.SetPushCacheHits(false);
  stripper.Strip(&primitives);

  indices.clear();
  for (unsigned int i=0; i<primitives.size(); ++i)
  {
    const triangle_stripper::primitive_group& p=primitives[i];

    switch (p.Type)
    {
      case triangle_stripper::TRIANGLES:
      {
        for (unsigned int j=0; j<p.Indices.size(); ++j)
          indices.push_back(p.Indices[j]);
        break;
      }

      case triangle_stripper::TRIANGLE_STRIP:
      {
        const triangle_stripper::indices& strip=p.Indices;
    
        for (int i=0, n=p.Indices.size()-2; i<n; ++i)
        {
          if (i%2)
          {
            indices.push_back( strip[i  ] );
            indices.push_back( strip[i+2] );
            indices.push_back( strip[i+1] );
          }
          else
          {
            indices.push_back( strip[i  ] );
            indices.push_back( strip[i+1] );
            indices.push_back( strip[i+2] );
          }
        }
        break;
      }
    }
  }
  
  unsigned int nT = indices.size()/3; 
  if (nT != n_fs_tris) 
  { 
    printf("Error while reordering indices with TriStripper. Aborting.\n");
    return;
  }
  //copy the results
  for (uint i = 0; i < n_fs_tris; ++i)
  {
    fs_tris_[i].v0 = indices[3*i  ];
    fs_tris_[i].v1 = indices[3*i+1];
    fs_tris_[i].v2 = indices[3*i+2];
  }
}

void TwoScaleModel::optimize_fs_indices(bool _create_strips)
{//optimization of triangle indices
  printf("*** Optimization of triangle indices:\n");
  printf("Pre-optimization visits statistics:\n");
  tr::TriangleReorder tri_reorder(const_cast<Triangle*>(&fs_tris_[0]), fs_tris_.size(), fs_vertices_.size(), 18);
  tri_reorder.calc_visits();

//   optimize_NvTriStrip(_create_strips);
//  tri_reorder.reorder();
  optimize_TriStripper();
  reorder_fs_vertices();
  
  printf("Post-optimization visits statistics:\n");
  tri_reorder.calc_visits();
}

void TwoScaleModel::reorder_fs_vertices()
{
  std::vector<int> vertices_map(fs_vertices_.size(), -1);
  //it is more convinient here to operate on indices than on triangles
  uint* idx_it = &fs_tris_[0].v0;
  uint next_idx = 0;
  for (uint i = 0, n = 3*fs_tris_.size(); i < n; ++i, ++idx_it)
  {//compute the mapping of the vertices 
    if (vertices_map[*idx_it] < 0)
    {
      vertices_map[*idx_it] = next_idx++;
    }
  }  
  CHECK(next_idx == fs_vertices_.size());
  
  idx_it = &fs_tris_[0].v0;
  for (uint i = 0, n = 3*fs_tris_.size(); i < n; ++i, ++idx_it)
  {//now correct the indices of the triangles
    CHECK(0 <= vertices_map[*idx_it] && vertices_map[*idx_it] < (int)fs_vertices_.size());
    *idx_it = vertices_map[*idx_it];
  }
  
  FineScaleVertexArray reord_fs_vertices(fs_vertices_.size());
  for (uint i = 0, n = fs_vertices_.size(); i < n; ++i)
  {//finally order vertex information
    reord_fs_vertices[vertices_map[i]] = fs_vertices_[i];
  }  
  
  fs_vertices_ = reord_fs_vertices;
}

void TwoScaleModel::clear_prim_groups()
{
  if (prim_groups_ != NULL)
  {
    delete [] prim_groups_;
    prim_groups_ = NULL;
    n_prim_groups_ = 0;
  }
}


/// TwoScaleModeling::CVTransform
class TwoScaleModeling::CVTransform
{
protected:
  TwoScaleModel*                          model_;
  uint                                    cv_id_;
  Vec3                                    cv_orig_pos_;
  Vec3                                    modif_vec_;
  float                                   step_modif_len_;
  float                                   max_modif_len_;
  float                                   curr_modif_len_;
  
public:
  CVTransform(TwoScaleModel& _model, uint _cv_id, const Vec3& _modif_vec, 
              float _step_modif_len, float _max_modif_len);
  ~CVTransform();
  void                                    perform();
  inline uint                             cv_id() const
  { return cv_id_; }
};

TwoScaleModeling::CVTransform::CVTransform(TwoScaleModel& _model, uint _cv_id, const Vec3& _modif_vec, 
                                           float _step_modif_len, float _max_modif_len)
: model_(&_model), cv_id_(_cv_id), cv_orig_pos_(model_->cm_vertices_[cv_id_]), modif_vec_(_modif_vec), 
  step_modif_len_(_step_modif_len), max_modif_len_(_max_modif_len), curr_modif_len_(0)
{
  CHECK(0 <= step_modif_len_ && step_modif_len_ < max_modif_len_);
  CHECK(cv_id_ < model_->cm_vertices_.size());
}

TwoScaleModeling::CVTransform::~CVTransform()
{ model_->cm_vertices_[cv_id_] = cv_orig_pos_; }//restore the original control point position

void TwoScaleModeling::CVTransform::perform()
{
  if ((step_modif_len_ > 0 && curr_modif_len_ < max_modif_len_) ||
      (step_modif_len_ < 0 && curr_modif_len_ > -max_modif_len_))
  {
    model_->cm_vertices_[cv_id_].add_linear(step_modif_len_, modif_vec_); 
    curr_modif_len_ += step_modif_len_;
  }
  else
  {
    step_modif_len_ *= -1;
  }
}

/// TwoScaleModeling


TwoScaleModeling::TwoScaleModeling(TwoScaleModel& _model)
: model_(_model)
{}

void TwoScaleModeling::set_modified_number(uint _n_modified)
{
  htf::set_min(_n_modified, model_.cm_vertices_.size());
  while (_n_modified > cv_transforms_.size())
  {
    inc_modified_number();
  }  
  while (_n_modified < cv_transforms_.size())
  {
    dec_modified_number();
  }
}

void TwoScaleModeling::add_modified_vertex(uint _vert_id)
{
  CHECK(_vert_id < model_.cm_vertices_.size());
  if (model_.cv_modified_[_vert_id])
  {//already modified
    return;
  }
  int axis_id = (cv_transforms_.size())%3;
  Vec3 modif_vec(0);
  modif_vec[axis_id] = 1;//unit vector at the one of the axes
  float step_modif_len = (float)(model_.cm_bb_.axis_length(axis_id)*0.01);
  float max_modif_len = step_modif_len*10.0f;
  
  model_.cv_modified_[_vert_id] = true;
  cv_transforms_.push_back(CVTransform(model_, _vert_id, modif_vec, step_modif_len, max_modif_len));
}

void TwoScaleModeling::inc_modified_number()
{
  if (cv_transforms_.size() < model_.cv_modified_.size())
  {
    add_modified_vertex(cv_transforms_.size());
  }
}

void TwoScaleModeling::dec_modified_number()
{
  if (cv_transforms_.size() > 0)
  {
    model_.cv_modified_[cv_transforms_.back().cv_id()] = false;
    cv_transforms_.pop_back();
    
    CHECK(cv_transforms_.size() < model_.cv_modified_.size());
  }
}

void TwoScaleModeling::shuffle_modified()
{
  uint n_modified = cv_transforms_.size();
  CHECK(n_modified <= model_.cv_modified_.size());
  
  cv_transforms_.clear();//clear all modified vertices 
  model_.cv_modified_.assign(model_.cv_modified_.size(), false);
  
  while (cv_transforms_.size() < n_modified)
  {
    uint cv_id = (uint)floor(model_.cm_vertices_.size()*(double)rand()/(double)RAND_MAX);
    add_modified_vertex(cv_id);
  }
}

void TwoScaleModeling::transform_cvs()
{
  for (CVTransformArray::iterator cvt_it = cv_transforms_.begin(); cvt_it != cv_transforms_.end(); ++cvt_it)
  {
    cvt_it->perform();
  }
}

uint TwoScaleModeling::n_cv_transforms() const
{ return cv_transforms_.size(); }
