/******************************************************************************
                          C++ Interface: TwoScaleModel
                            -------------------
    description       :
    begin             : Wed Mar 23 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#ifndef __TwoScaleModel_h__
#define __TwoScaleModel_h__

#include <math.h>
#include <float.h>
#include <vector>
#include <map>

#include "defs.h"
#include "Vector.h"
#include "../TriangleReorder/Triangle.h"
#include "../NvTriStrip/NvTriStrip.h"

using tr::Triangle;

struct ControlVertexCoeff
{
  uint                                      cv_id_;
  float                                     f_;
  float                                     du_;
  float                                     dv_;
};

struct FineScaleVertex
{
  typedef std::vector<ControlVertexCoeff>   ControlVertexCoeffArray;
  
  float                                     pt_displ_;//normal displacement
  float                                     t0_displ_n_, t0_displ_du_, t0_displ_dv_;
  float                                     t1_displ_n_, t1_displ_du_, t1_displ_dv_;
  ControlVertexCoeffArray                   cv_coeffs_;//coefficients of related control vertices
};

struct BoundingBox
{
  typedef Vec3::Scalar                      Scalar;
  Vec3                                      corners_[2];

  BoundingBox() {}
  template <class _Iterator>
  BoundingBox(_Iterator _begin, _Iterator _end) { build(_begin, _end); }

  inline Scalar                             axis_length(uint _axis) const
  {
    CHECK(_axis < 3);
    return corners_[1][_axis] - corners_[0][_axis];
  }

  inline Scalar                             diagonal_length() const
  {
    Vec3 diag(corners_[0]);
    diag -= corners_[1];
    return diag.norm();
  }

  inline Scalar                             volume() const
  {
    return (corners_[1][0] - corners_[0][0])*(corners_[1][1] - corners_[0][1])*(corners_[1][2] - corners_[0][2]);
  }

  void                                      invalidate()
  {
    corners_[0] = Vec3(FLT_MAX);
    corners_[1] = Vec3(-FLT_MAX);
  }

  __inline__ bool                           is_valid() const
  {
    return  corners_[0][0] <= corners_[1][0] && 
            corners_[0][1] <= corners_[1][1] && corners_[0][2] <= corners_[1][2];
  }

  template <class _Point>
  __inline__ void                           update(const _Point& _pt)
  {
    htf::set_min(corners_[0][0], _pt[0]);
    htf::set_min(corners_[0][1], _pt[1]);
    htf::set_min(corners_[0][2], _pt[2]);
    htf::set_max(corners_[1][0], _pt[0]);
    htf::set_max(corners_[1][1], _pt[1]);
    htf::set_max(corners_[1][2], _pt[2]);
  }

  template <class _Iterator>
  void                                      build(_Iterator _begin, _Iterator _end)
  {
    invalidate();
    for (; _begin != _end; ++_begin)
    {
      update(*_begin);
    }
  }
  template <class _Point>
  bool                                      is_in(const _Point& _pt) const
  {
    return corners_[0][0] <= _pt[0] && _pt[0] <= corners_[1][0] &&
           corners_[0][1] <= _pt[1] && _pt[1] <= corners_[1][1] &&
           corners_[0][2] <= _pt[2] && _pt[2] <= corners_[1][2];
  }
  
  void                                      calc_mid_pt(Vec3& _mid_pt) const
  {
    _mid_pt.set_sum(corners_[0], corners_[1]);
    _mid_pt /= 2;
  }
};


struct TwoScaleModel
{
  typedef std::vector<Vec3>                 Vec3Array;
  typedef std::vector<FineScaleVertex>      FineScaleVertexArray;
  typedef std::vector<Triangle>             TriangleArray;
  typedef std::vector<bool>                 FlagArray;
  typedef std::vector<Vec3Array>            CVCoeffArray;
  
  Vec3Array                                 cm_vertices_;
  FlagArray                                 cv_modified_;
  BoundingBox                               cm_bb_;//bounding box of the control mesh vertices
  FineScaleVertexArray                      fs_vertices_;
  TriangleArray                             fs_tris_;
  
  PrimitiveGroup*                           prim_groups_;
  uint                                      n_prim_groups_;

  TwoScaleModel();
  ~TwoScaleModel();
    
  bool                                      load(const char* _file_name);
  void                                      update_cm_bb()
  { cm_bb_.build(cm_vertices_.begin(), cm_vertices_.end()); }
  
  void                                      dump_stats() const;
  /** calculates completely the fine scale geometry */
  void                                      calc_geometry(Vec3* _vertices,  Vec3* _normals) const;
  /** returns the displacements in interleaved format. */
  void                                      get_displacements(Vec3Wt* _displ) const;
  /** Splits the fine scale geometry into two parts: the static part consists of the linear combinations
      of the static control vertices for every fine scale vertiex and is written the arrays _foot_*[]. 
      The dynamic part (_cv_coeff) is the coefficients of the linear combination required to compute 
      the contribution of the modified control points to the _foot_*[] values. Only the coefficients 
      arrays cooresponding to modified control vertices are created. All other arrays are empty().
      Returns the number of modifed control vertices (= non-empty control vertices coefficients arrays).
  */
  uint                                      calc_geometry_split(Vec3* _foot_pt, Vec3* _foot_du, Vec3* _foot_dv,
                                                                CVCoeffArray& _cv_coeff) const;
  
  void                                      optimize_fs_indices(bool _create_strips = false);
  void                                      clear_prim_groups();
  void                                      optimize_NvTriStrip(bool _create_strips);
  void                                      optimize_TriStripper();
  void                                      reorder_fs_vertices();
};


class TwoScaleModeling 
{
private:
  class CVTransform;
  
  typedef std::vector<CVTransform>          CVTransformArray;
  
  TwoScaleModel&                            model_;
  CVTransformArray                          cv_transforms_;
  
public:
  TwoScaleModeling(TwoScaleModel& _model);
  
  void                                      set_modified_number(uint _n_modifed);
  void                                      add_modified_vertex(uint _vert_id);
  void                                      inc_modified_number();
  void                                      dec_modified_number();
  void                                      shuffle_modified();
  
  void                                      transform_cvs();
  uint                                      n_cv_transforms() const;
};
#endif//__TwoScaleModel_h__
