/******************************************************************************
                          C++ Implementation: tsrb
                            -------------------
    description       :
    begin             : Tue Mar 22 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt in_file that comes with this distribution
 ******************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <vector>
#include <string>

#include <GL/glew.h>
#include <GL/glut.h>

#include "TwoScaleModel.h"
#include "IVisualization.h"
#include "Timer.h"
#include "GLTrackball.hh"

#ifdef WIN32
#define snprintf _snprintf
#endif

#define OGL_MAX_X                 10000
#define OGL_MIN_X                 - OGL_MAX_X
#define OGL_MAX_Y                 10000
#define OGL_MIN_Y                 - OGL_MAX_Y
#define OGL_MAX_Z                 10000
#define OGL_MIN_Z                 - OGL_MAX_Z

#define OGL_PERSPECTIVE_ANGLE     45.0
#define OGL_CLIP_NEAR             2
#define OGL_CLIP_FAR              2*OGL_MAX_Z
#define OGL_CLEAR_DEPTH           1.0

/// VisualizationFactory 
inline KeyVisualizationMap& VisualizationFactory::instance()
{
  static KeyVisualizationMap map__;
  return map__;
}
const KeyVisualizationMap& VisualizationFactory::map()
{ return instance(); }
void VisualizationFactory::connect(char _key, VisualizationProducer _vp_f)
{ instance()[_key] = _vp_f; }


//input parameterers
static int argc__ = 0;
static char** argv__ = NULL;
//model
static TwoScaleModel ts_model__; 
static TwoScaleModeling* ts_modeling__ = NULL;
//visualization
static IVisualization* vis__ = NULL;
static char curr_vis_key__ = '1';
static IVisualization::ModeEnum vis_mode__ = IVisualization::MODE_SOLID;
//misc
static Timer timer__;
static double z_zoom__ = 4;  
static bool rotate__ = false;
static ACG::GLTrackball* trackball__ = NULL;


void clean_and_exit(int _exit_code)
{
  delete vis__;
  exit(_exit_code);
}

void print_usage_and_exit(int _exit_code = 0)
{
  printf("Usage:\n");
  printf("tsrb [-h] [-i] [-n x] -f two_scale_file_name\n");
  clean_and_exit(_exit_code);
}

void optimize_indices_early(bool _create_strips)
{
  int time_start = glutGet(GLUT_ELAPSED_TIME);
  ts_model__.optimize_fs_indices(_create_strips);
  int time_end = glutGet(GLUT_ELAPSED_TIME);
  double elapsed_sec = (time_end - time_start)/1000.0;
  printf("Elapsed time for the reordering: %0.2fsec. Reseting visualization(s)\n", elapsed_sec);
}

void init(int argc, char *argv[])
{
  bool loaded = false;
  bool tnl_cache_optim_enabled = true;

  for (int i=1; i < argc; ++i)
  {
    if (strcmp(argv[i], "-i") == 0) 
    {
      printf("GL_RENDERER   = %s\n", (char*) glGetString(GL_RENDERER));
      printf("GL_VERSION    = %s\n", (char*) glGetString(GL_VERSION));
      printf("GL_VENDOR     = %s\n", (char*) glGetString(GL_VENDOR));
      printf("GL_EXTENSIONS = %s\n", (char*) glGetString(GL_EXTENSIONS));
    }
    else if (strcmp(argv[i], "-f") == 0 && (i + 1) < argc)
    {//load two scale model file
      printf("Loading two scale model file: %s ...", argv[i+1]);
      if (!ts_model__.load(argv[i+1]))
      {
        printf(" failed!\n");
        print_usage_and_exit(1);
      }
      printf(" loaded.\n");
      ts_model__.dump_stats();
      timer__.set_scene_complexity(ts_model__.fs_vertices_.size(), ts_model__.fs_tris_.size());
      ts_modeling__ = new TwoScaleModeling(ts_model__);
      loaded = true;
    }
    else if (strcmp(argv[i], "-D") == 0)
    {
      tnl_cache_optim_enabled = false;
    }
    else if (strcmp(argv[i], "-h") == 0)
    {
      print_usage_and_exit();
    }
  }
  if (!loaded)
  {
    print_usage_and_exit(1);
  }
  if (tnl_cache_optim_enabled)
  {
    printf("Optimizing fine scale mesh w.r.t. the T&L post-transform cache...\n");
    printf("Use -D to disable this optimization (not recommended).\n");
    optimize_indices_early(false);
  }
  else
  {
    printf("Optimizing fine scale mesh w.r.t. the T&L post-transform cache DISABLED\n");
    printf("Perfomance might be suboptimal. Press 'o' at runtime to invoke the optimization.\n");
  }
}

void query_limits()
{
  int max_vertex_attribs;
  glGetIntegerv(GL_MAX_VERTEX_ATTRIBS_ARB, &max_vertex_attribs);
  printf("Maximum vertex attributes: %d.\n", max_vertex_attribs);
  int max_vertex_uniform_components;
  glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB, &max_vertex_uniform_components);
  printf("Maximum vertex uniform components: %d.\n", max_vertex_uniform_components);
  int max_vertex_texture_units;
  glGetIntegerv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB, &max_vertex_texture_units);
  printf("Maximum vertex texture units: %d.\n", max_vertex_texture_units);
  int max_texture_units;
  glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &max_texture_units);
  printf("Maximum texture units: %d.\n", max_texture_units);
  int max_texture_size;
  glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_texture_size);
  printf("Maximum texture size: %d.\n", max_texture_size);
}

void init_OpenGL()
{
  GLenum err = glewInit();
  if (GLEW_OK != err)
  {
    /* Problem: glewInit failed, something is seriously wrong. */
    printf("GLEW Error: %s\n", glewGetErrorString(err));
    clean_and_exit(1);
  }
  //check for some extensions 
  if (!GLEW_ARB_vertex_buffer_object)
  {
    printf("Error: ARB_vertex_buffer_object not supported. Aborting.\n");
    clean_and_exit(1);
  }
  if (!GLEW_ARB_vertex_program)
  {
    printf("Error: ARB_vertex_program not supported. You won't be able to run the GPU rendering code.\n");
  }
  
  OGL_CHECK(::glDisable(GL_CULL_FACE));
  OGL_CHECK(::glDisable(GL_LOGIC_OP));
  
  // depth settings  
  OGL_CHECK(::glEnable(GL_DEPTH_TEST));
  OGL_CHECK(::glDepthFunc(GL_LEQUAL));
  OGL_CHECK(::glClearDepth(OGL_CLEAR_DEPTH));
  OGL_CHECK(::glClearColor(0, 0, 0, 0));
  
  // lighting & material settings
  float mat_specular[4] = {0.5, 0.5, 0.5, 1};
  float mat_emission[4] = {0.1, 0.1, 0.1, 1};//fix - export
  
  OGL_CHECK(::glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular));
  OGL_CHECK(::glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, mat_emission));
  OGL_CHECK(::glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128));
  OGL_CHECK(::glEnable(GL_COLOR_MATERIAL));
  OGL_CHECK(::glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE));
  OGL_CHECK(::glColor3f(0.8, 0.6, 0.0));
  
  float ambient_col[4] = {0, 0, 0, 0};
  float diffuse_col[4] = {1, 1, 1, 1};
  float specular_col[4] = {0, 0, 0, 0};
  float position[4] = { 0, 1, 1, 0};
  
  OGL_CHECK(::glLightfv(GL_LIGHT0, GL_AMBIENT, ambient_col));
  OGL_CHECK(::glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse_col));
  OGL_CHECK(::glLightfv(GL_LIGHT0, GL_SPECULAR, specular_col));
  OGL_CHECK(::glLightfv(GL_LIGHT0, GL_POSITION, position) );
  
  OGL_CHECK(::glEnable(GL_NORMALIZE));
  OGL_CHECK(::glShadeModel(GL_SMOOTH));
//   OGL_CHECK(::glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE));
//   OGL_CHECK(::glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE));
  OGL_CHECK(::glEnable(GL_LIGHTING));
  OGL_CHECK(::glEnable(GL_LIGHT0));
  
  query_limits();
}

void set_projection_matrix()
{
//   float scale_factor = diag_len > 0 ? 1.0/diag_len : 1;//fit the object into 50% of the cube [-1:1]
    
  uint w = glutGet(GLUT_WINDOW_WIDTH);
  uint h = glutGet(GLUT_WINDOW_HEIGHT);
  
  OGL_CHECK(::glViewport(0, 0, w, h));
  OGL_CHECK(::glMatrixMode(GL_PROJECTION));
  OGL_CHECK(::glLoadIdentity());
  
  double aspect = h != 0 ? (double)w/(double)h : 1.0;
  double z_range = ts_model__.cm_bb_.diagonal_length();
  double near_z = 1.0;
  double far_z = z_zoom__*z_range + near_z;
  
//   OGL_CHECK(::gluPerspective(45.0, aspect, 1.0, 2000.0));
  OGL_CHECK(::gluPerspective(45.0, aspect, near_z, far_z));
  
  OGL_CHECK(::glMatrixMode(GL_MODELVIEW));
  OGL_CHECK(::glLoadIdentity());

  Vec3 transl(0, 0, -far_z + z_range/2.0);//target pos is between the near and far clipping planes
  Vec3 mid_pt;
  ts_model__.cm_bb_.calc_mid_pt(mid_pt);
  transl -= mid_pt;
  OGL_CHECK(::glTranslatef(transl.x, transl.y, transl.z));
}

void init_visualization(char _key)
{
  delete vis__;
  vis__ = NULL;
  
  ts_model__.update_cm_bb();
  set_projection_matrix();
  
  KeyVisualizationMap::const_iterator vf_it = VisualizationFactory::map().find(_key);
  if (vf_it != VisualizationFactory::map().end())
  {//there is a visualization producer associated with this key - try it
    IVisualization* new_vis = (*vf_it->second)(ts_model__, argc__, argv__);
    if (new_vis == NULL || !new_vis->init())
    {
      delete new_vis;
      printf("Selected visualization failed to initialize. Reverting...\n");
      if (_key != curr_vis_key__)
      {
        init_visualization(curr_vis_key__);
        return;
      }
    }
    else
    {
      vis__ = new_vis;
      curr_vis_key__ = _key;
      printf("Visualization with key %c initialized successfuly.\n", _key);
    }
  }
  if (vis__ == NULL)
  {//_key visualization failed to create/initialize ...
    printf("Looking for a \"good\" visualization out of %d possible... \n", 
            VisualizationFactory::map().size());
    for (KeyVisualizationMap::const_iterator vf_it = VisualizationFactory::map().begin(),
                                             vf_end = VisualizationFactory::map().end(); 
         vf_it != vf_end; ++vf_it)
    {
      printf("Attempting visualization with key %c ... ", vf_it->first);
      if (vf_it->first == _key)
      {//this key was tried already...
        printf(" already tried.\n");
        continue;
      }
      IVisualization* new_vis = (*vf_it->second)(ts_model__, argc__, argv__);
      if (new_vis == NULL || !new_vis->init())
      {
        printf("failed.\n");
        delete new_vis;
      }
      else
      {
        printf("success.\n");
        vis__ = new_vis;
        curr_vis_key__ = _key;
        break;
      }
    }
  }
  if (vis__ == NULL)
  {//all visualizations failed - exit here
    printf("All visualizations failed. Aborting.\n");
    clean_and_exit(1);
  }
  timer__.reset();  
}

void inc_modified_number()
{
  ts_modeling__->inc_modified_number();
  printf("Currently modifying %d vertices.\n", ts_modeling__->n_cv_transforms());
  init_visualization(curr_vis_key__);
}

void dec_modified_number()
{
  ts_modeling__->dec_modified_number();
  printf("Currently modifying %d vertices.\n", ts_modeling__->n_cv_transforms());
  init_visualization(curr_vis_key__);
}

void shuffle_modified()
{
  ts_modeling__->shuffle_modified();
  printf("Currently modifying %d random vertices.\n", ts_modeling__->n_cv_transforms());
  init_visualization(curr_vis_key__);
}

void optimize_indices(bool _create_strips)
{
  optimize_indices_early(_create_strips);
  init_visualization(curr_vis_key__);
}

/// GLUT handlers

static void draw()
{
  OGL_CHECK(glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
  
  ts_modeling__->transform_cvs();
  
  OGL_CHECK(::glColor3f(0.8, 0.6, 0.0));
  OGL_CHECK(::glPushMatrix());
  
  if (rotate__)
  {
    GLint t = glutGet(GLUT_ELAPSED_TIME);
    uint rdps = (uint)floor(t*150.0/1000.0) % 360;
//     glRotated(rdps, 0, 0, 1);
    glRotated(rdps, 0, 1, 0);
    glRotated(rdps, 1, 0, 0);
  }
  
  vis__->draw(vis_mode__);
  
  OGL_CHECK(::glPopMatrix());

//   OGL_CHECK(::glPushMatrix());
//   OGL_CHECK(::glLoadIdentity());
  
/*  OGL_CHECK(::glColor3f(1, 1, 1));
  OGL_CHECK(::glRasterPos2i(0,0));
  for (const char* s = buffer; *s; s++)
  {
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *s);
  }*/
  
//   OGL_CHECK(::glPopMatrix());
  
  glutSwapBuffers();
  
  double fps, mvps, mtps;
  if (timer__.update(fps, mvps, mtps))
  {
    char buffer[1024];//Two Scale Rendering Benchmark: 
    snprintf(buffer, 1024, "FPS: %.3f / MVPS: %.3f / MTPS: %.3f", fps, mvps, mtps);
    glutSetWindowTitle(buffer);
  }
}

static void idle(void)
{ glutPostRedisplay(); }

static void reshape(int /*width*/, int /*height*/)
{ set_projection_matrix(); }

static void special(int k, int /*x*/, int /*y*/)
{}

static void key(unsigned char k, int /*x*/, int /*y*/)
{
  timer__.reset();
  
  switch (k) 
  {
  case 'P'  :
  case 'p'  : 
    glPolygonMode(GL_FRONT_AND_BACK, GL_POINT); 
    vis_mode__ = IVisualization::MODE_POINTS;
    break;
  case 'W'  :
  case 'w'  : 
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); 
    vis_mode__ = IVisualization::MODE_WIREFRAME;
    break;
  case 'S'  :
  case 's'  : 
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); 
    vis_mode__ = IVisualization::MODE_SOLID;
    break;
  case 'Z'  :
  case 'z'  : z_zoom__ *= 0.9; set_projection_matrix(); break;
  case 'X'  :
  case 'x'  : z_zoom__ *= 1.2; set_projection_matrix(); break;
  case 'R'  :
  case 'r'  : rotate__ = !rotate__; break;
  case 'T'  :
  case 't'  : optimize_indices(true); break;
  case 'O'  :
  case 'o'  : optimize_indices(false); break;
  case '='  :
  case '+'  : inc_modified_number(); break;
  case '-'  :
  case '_'  : dec_modified_number(); break;
  case 'D'  :
  case 'd'  : vis__->dump(); timer__.reset(); break;
  case 'F'  :
  case 'f'  : shuffle_modified(); break;
  case 27   : exit(0); break;
  default:
    if (VisualizationFactory::map().find(k) != VisualizationFactory::map().end())
    {
      init_visualization(k);
    }
  }
  glutPostRedisplay();
}

static void visible(int vis)
{
  if (vis == GLUT_VISIBLE)
    glutIdleFunc(idle);
  else
    glutIdleFunc(NULL);
}

static void mouse(int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON)
  {
    if (state == GLUT_DOWN)
    {
      delete trackball__;
      trackball__ = new ACG::GLTrackball();
      
      Vec3 model_center;
      ts_model__.cm_bb_.calc_mid_pt(model_center);
      trackball__->set_center(model_center);
      trackball__->mouse_press(0, x, y);
    }
    else
    {
      trackball__->mouse_release(0, 0, 0);
      trackball__ = NULL;
    }
  }
}

static void motion(int x, int y)
{
  if (trackball__ != NULL)
  {
    trackball__->mouse_move(x, y);
  }   
}

int main(int argc, char *argv[])
{
  init(argc, argv);
  argc__ = argc;
  argv__ = argv;
  
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(800, 600);
  glutCreateWindow("Two Scale Rendering Benchmark");
  
  init_OpenGL();

  glutDisplayFunc(draw);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(key);
  glutSpecialFunc(special);
  glutVisibilityFunc(visible);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);  
  
  init_visualization(curr_vis_key__);  
    
  glutMainLoop();
  
  clean_and_exit(0);
  return 0;
}



//   glEnable(GL_CULL_FACE);
//   glCullFace(GL_BACK);
  
  
//   OGL_CHECK(::glDrawBuffer(GL_BACK));
/*  double max_tex_size, max_tex_units;
  ::glGetDoublev(GL_MAX_TEXTURE_UNITS_ARB, &max_tex_units);
  ::glGetDoublev(GL_MAX_TEXTURE_SIZE, &max_tex_size);
  printf("Max texture size: %f, max. texture units: %f\n", max_tex_size, max_tex_units);*/
